"""Manage the target data for Grape City (gc) alignments."""

from collections import UserDict
from csv import DictReader
from dataclasses import dataclass

from src import config

# these attribute names match the source data for simplicity


@dataclass
class Target:
    """Manage target data for a word."""

    # Identifies the word/morph in BBCCCVVVWWWP format
    # note, no word part: so this doesn't support sub-word tokens
    identifier: str
    # ?
    altId: str
    # surface form: omit this for copyrighted sources
    text: str
    # ?
    transType: str
    # is this token punctuation?
    # Not always set, at least not for YLT
    isPunc: bool
    # if the same word is aligned to multiple source words, the
    # primary word is the content-bearing term? Example from 42004003
    isPrimary: bool
    _fields: tuple = ("identifier", "altId", "text", "transType", "isPunc", "isPrimary")

    def __repr__(self) -> str:
        """Return a printed representation."""
        return f"<Target: {self.identifier}>"

    @staticmethod
    def fromrow(row) -> "Target":
        """Return an instance from a row of data."""
        row["isPunc"] = True if row["isPunc"] == "True" else False
        row["isPrimary"] = True if row["isPrimary"] == "True" else False
        return Target(**row)

    @property
    def token(self) -> str:
        """Return text.

        For compatability with Source()."""
        return self.text

    def display(self) -> None:
        """Print a readable display of the key data."""
        print(f"{self.identifier}: {self.text:<20} ('{self.transType}', {self.isPunc}, {self.isPrimary})")


class Reader(UserDict):
    """Manages target data read from file."""

    def __init__(self, sourceid: str, targetid: str) -> None:
        """Initialize Reader instance."""
        self.sourceid = sourceid
        self.targetid = targetid
        super().__init__(self)
        with (config.TARGETS / f"{sourceid}-{targetid}.tsv").open(encoding="utf-8") as f:
            dictreader = DictReader(f, fieldnames=Target._fields, delimiter="\t")
            self.data = {target.identifier: target for row in dictreader if (target := Target.fromrow(row))}



"""Process alignment data formatted by Grape City.

These alignments were then handed down to Global Bible Initiative (GBI), and then Clear Bible. Note that some of the target texts were typed in by hand, so may have minor differences from authoritative editions. 

The source alignment data is in JSON, with a list of objects
representing verse data. Each verse object has three keys:

- manuscript: a single key "words" whose value is a list of source
  text word/morph objects, with properties like "id", "text", "lemma",
  etc.
- translation: a single key "words" whose value is a list of target
  text word/morph objects, with properties like "id", "text",
  "isPunc", etc.
- links: a list of lists of lists, each terminal list containing one or more indexes into the manuscript and translation words (zero-based). So an element of links like [[0], [0, 1]] means the first word of the source text is aligned with the first two words of the target text.

"""


from dataclasses import dataclass
import json
from pathlib import Path
from typing import Union

#
ROOT = Path(__file__).parent.parent.parent
# these don't belong here
OTJSON = ROOT / "data/eng/ylt/ylt.ot.alignment.json"
NTJSON = ROOT / "data/eng/ylt/ylt.nt.alignment.json"

# type that's either None (undefined) or a float. This allows 0.0 as a
# declared value.
nonefloat = Union[str, None]

# these attribute names match the source data for simplicity


@dataclass
class Manuscript:
    """Manage manuscript data for a word/morph."""

    # Identifies the word/morph in BBCCCVVVWWWP format
    id: str
    # ?
    altId: str
    # surface form: omit this for copyrighted sources
    text: str
    # Strongs number
    strong: str
    # English gloss
    gloss: str
    # sometimes a Chinese/alternate language gloss?
    gloss2: str
    # source language lemma
    lemma: str
    # part of speech
    pos: str
    # coded morphological information: need to document the format
    morph: str

    # TODO: enumerate and validate part of speech values

    @staticmethod
    def fromjsondict(jdict: dict[str, str]) -> "Manuscript":
        """Return a Manuscript instance for a dictionary read from JSON."""
        # convert id attr to a str
        jdict["id"] = str(jdict["id"])
        if len(jdict["id"]) == 11:
            # add missing leading zero: fragile hack!
            jdict["id"] = "0" + jdict["id"]
        return Manuscript(**jdict)


@dataclass
class Translation:
    """Manage translation data for a word."""

    # Identifies the word/morph in BBCCCVVVWWWP format
    # note, no word part: so this doesn't support sub-word tokens
    id: str
    # ?
    altId: str
    # surface form: omit this for copyrighted sources
    text: str
    # ?
    transType: str
    # is this token punctuation?
    # Not always set, at least not for YLT
    isPunc: bool
    # is this token primary? Not sure what the semantics are.
    isPrimary: bool

    @staticmethod
    def fromjsondict(jdict: dict[str, str]) -> "Translation":
        """Return a Manuscript instance for a dictionary read from JSON."""
        # convert id attr to a str
        jdict["id"] = str(jdict["id"])
        if len(jdict["id"]) == 10:
            # add missing leading zero: fragile hack!
            jdict["id"] = "0" + jdict["id"]
        return Translation(**jdict)


@dataclass
class AlignmentGroup:
    """Manage alignment data for a word.

    Zero or more items, typically 1, or 2 if many-to-many. The values
    are indexes into the sequences of sourceitems and targetitems.

    Confidence defaults to undefined (None): if provided, it should be
    a float from 0.0-1.0 inclusive. Not provided for manual alignments
    (no assumption that human aligners are perfect).

    """

    sourceitems: tuple
    targetitems: tuple
    # empty by default
    confidence: nonefloat = None

    def __repr__(self) -> str:
        """Return a printed representation."""
        return f"<AGroup: [{self.sourceitems}, {self.targetitems}]>"


class VerseData:
    """Manages data for a single verse.

    alignmentgroups values are indexes into source/targetitems, so order matters.
    """

    verseid: str = ""
    sourceitems: list = []
    targetitems: list = []
    alignmentgroups: list = []
    # empty by default
    confidence: nonefloat = None

    def __init__(self, versedata: dict[str, dict]):
        """Initialize VerseData instance."""
        self.sourceitems = [Manuscript.fromjsondict(w) for w in versedata["manuscript"]["words"]]
        self.targetitems = [Translation.fromjsondict(w) for w in versedata["translation"]["words"]]
        # need more logic here for handling confidence scores
        self.alignmentgroups = [AlignmentGroup(ag[0], ag[1]) for ag in versedata["links"]]
        self.verseid = self.sourceitems[0].id[:8]

    def __repr__(self) -> str:
        """Return a printed representation."""
        return f"<VerseData({self.verseid})>"

    def aligned_items(self, itemsattr: "targetitems") -> list:
        """Return a list of source or target items that are in alignment groups."""
        assert itemsattr in ["sourceitems", "targetitems"], f"Invalid itemsattr value: {itemsattr}"
        included = sorted([ti for ag in self.alignmentgroups for ti in getattr(ag, itemsattr)])
        return [getattr(self, itemsattr)[titem] for titem in included]

    # possible extensions
    # - display unaligned source/target items
    # - validate no double grouping
    # - default order seems to be source sequence: might want target sequence too


class Reader:
    """Manages data read from Grape City JSON."""

    verseitems: list = []

    def __init__(self, jsonfile: Path):
        """Initialize Reader instance."""
        with jsonfile.open(encoding="utf-8") as f:
            self.verseitems = [VerseData(v) for v in json.load(f)]



"""Generate a catalog of alignments.

>>> from src import catalog
>>> catalog.Catalog().write()

TODO:
- add check for missing source/target files

"""

from copy import deepcopy
from csv import DictWriter
from functools import reduce
from pathlib import Path
from warnings import warn

import tomli

from src import config


class Catalog:
    """Manage data across all the alignments"""

    alignments: Path = config.ALIGNMENTS
    catalogpath: Path = config.DATAPATH / "catalog.tsv"
    # Standard metadata attributes: warn if not present
    stdattrs: dict[str, dict[str, str]] = {
        "alignment": ["format", "identifier", "license", "process", "scope", "team"],
        "source": ["identifier", "license"],
        "target": ["identifier", "license", "name", "url"],
    }
    # no warnings, but don't include in output
    omittedattrs: dict[str, dict[str, str]] = {
        "alignment": [
            # already in the key
            "identifier",
            # aliready part of the identifier
            "process",
        ],
        "source": [
            # already in the key
            "identifier"
        ],
        "target": [
            # already in the key
            "identifier"
        ],
    }

    def __init__(self) -> None:
        """Initialize an instance."""
        self.languages = sorted([l.name for l in self.alignments.glob("*")])
        self.versions = sorted([(lang, v.name) for lang in self.languages for v in self.alignments.glob(f"{lang}/*")])
        self.tomlfiles = {
            # TODO: tomlfile.stem is sufficient to identify an
            # alignment. Maybe leave these three elements separate
            # therefore?
            f"{lang}+{version}+{tomlfile.stem}": tomlfile
            for lang, version in self.versions
            for tomlfile in sorted(self.alignments.glob(f"{lang}/{version}/*.toml"))
        }
        self.tomldicts = {}
        for alignedver, tomlfile in self.tomlfiles.items():
            with tomlfile.open("rb") as f:
                self.tomldicts[alignedver] = tomli.load(f)
        self.commonkeys = {k for v in self.tomldicts.values() for k in v}

    def write(self) -> None:
        """Write the catalog."""
        langverkey = "lang+version+alignment"
        self.langverdicts = {}
        for alignedver, tomldict in self.tomldicts.items():
            # warn if standard attrs are missing
            self._validate(alignedver, tomldict)
            stddict = deepcopy(tomldict)
            # drop non-standard pairs
            for stdk in stddict.copy():
                if stdk not in self.stdattrs:
                    warn(f"Dropping {stdk} from {alignedver} data: non-standard.")
                    del stddict[stdk]
                else:
                    for stdsubk in stddict[stdk].copy():
                        if stdsubk not in self.stdattrs[stdk]:
                            warn(f"Dropping {stdk}.{stdsubk} from {alignedver}: non-standard.")
                            del stddict[stdk][stdsubk]
                        elif stdsubk in self.omittedattrs[stdk]:
                            del stddict[stdk][stdsubk]
            self.langverdicts[alignedver] = {
                f"{k}.{subk}": stddict[k][subk] if subk in stddict[k] else ""
                for k in self.stdattrs
                for subk in self.stdattrs[k]
                if subk not in self.omittedattrs[k]
            }
            # reformat name
            fixeddict = self.langverdicts[alignedver]
            if "target.name" in fixeddict and isinstance(fixeddict["target.name"], dict):
                langcode, langname = list(fixeddict["target.name"].items())[0]
                fixeddict["target.name"] = f"'{langname}'@{langcode}"
        self.fieldnames = [langverkey] + [
            f"{k}.{subk}" for k in self.stdattrs for subk in self.stdattrs[k] if subk not in self.omittedattrs[k]
        ]
        with self.catalogpath.open("w") as f:
            writer = DictWriter(f, fieldnames=self.fieldnames, delimiter="\t")
            writer.writeheader()
            for alignedver, langverdict in self.langverdicts.items():
                # langvervalue = f"{alignedver[0]}+{alignedver[1]}"
                langverdict.update({langverkey: alignedver})
                writer.writerow(langverdict)

    # TODO: add as_markdown() to output a table
    def _validate(self, langver: str, langverdict: dict[str, dict[str, str]]) -> None:
        """Warn if standard attributes are missing."""
        for stdk in self.stdattrs:
            if stdk not in langverdict:
                warn(f"{langver} is missing standard key '{stdk}'")
            for stdsubk in self.stdattrs[stdk]:
                if stdsubk not in langverdict[stdk]:
                    warn(f"{langver} is missing standard subkey {stdsubk}")



"""Manage the source data for Grape City (gc) alignments.

>>> from src import gcsource
# read NA/27 GNT data for the LEB alignment
>>> rd = gcsource.Reader(sourceid="NA27", targetid="LEB")
"""

from collections import UserDict
from csv import DictReader
from dataclasses import dataclass

from src import config

# these attribute names match the source data for simplicity


@dataclass
class Source:
    """Manage source data for a word/morph."""

    # Identifies the word/morph in BBCCCVVVWWWP format
    identifier: str
    # ?
    altId: str
    # surface form: omit this for copyrighted sources
    text: str
    # Strongs number
    strongs: str
    # English gloss
    gloss: str
    # sometimes a Chinese/alternate language gloss?
    gloss2: str
    # source language lemma
    lemma: str
    # part of speech
    pos: str
    # coded morphological information: need to document the format
    morph: str
    _missing: str = "--"
    _fields: tuple = ("identifier", "altId", "text", "strongs", "gloss", "gloss2", "lemma", "pos", "morph")
    # TODO: enumerate and validate part of speech values

    def __repr__(self) -> str:
        """Return a printed representation."""
        return f"<Source: {self.identifier}>"

    @property
    def token(self) -> str:
        """Return text if available, else lemma."""
        return self.lemma if (self.text == self._missing) else self.text

    def display(self) -> None:
        """Print a readable display of the key data."""
        print(f"{self.identifier}: {self.text:<20} ('{self.gloss}', '{self.lemma}', {self.pos})")


class Reader(UserDict):
    """Manages source data read from file."""

    def __init__(self, sourceid: str, targetid: str) -> None:
        """Initialize Reader instance."""
        self.sourceid = sourceid
        self.targetid = targetid
        super().__init__(self)
        with (config.SOURCES / f"{sourceid}-{targetid}.tsv").open(encoding="utf-8") as f:
            dictreader = DictReader(f, fieldnames=Source._fields, delimiter="\t")
            self.data = {source.identifier: source for row in dictreader if (source := Source(**row))}



"""Initialize common values."""

from pathlib import Path

ROOT = Path(__file__).parent.parent
DATAPATH = ROOT / "data"

ALIGNMENTS = DATAPATH / "alignments"
SOURCES = DATAPATH / "sources"
TARGETS = DATAPATH / "targets"
NAMES = DATAPATH / "names"



"""Read alignment data from Grape City.

Note that some of the target texts were typed in by hand, so may have
minor differences from authoritative editions.

"""

from collections import UserDict
from dataclasses import dataclass, field
import itertools
import json
from typing import Union

from src import config, gcsource, gctarget

# from . import config, gcsource, gctarget

# type that's either None (undefined) or a float. This allows 0.0 as a
# declared value.
nonefloat = Union[str, None]


@dataclass
class AlignmentGroup:
    """Manage alignment data for a word.

    sourceitems and targetitems are sequences of zero or more items,
    typically 1, or 2 if many-to-many. The values are instances of Source  indexes into the
    sequences of sourceitems and targetitems.

    Confidence defaults to undefined (None): if provided, it should be
    a float from 0.0-1.0 inclusive. Not provided for manual alignments
    (no assumption that human aligners are perfect).

    """

    identifier: str
    sourceitems: tuple
    targetitems: tuple
    # empty by default
    meta: dict = field(default_factory=dict)

    def __repr__(self) -> str:
        """Return a printed representation."""
        return f"<AGroup({self.identifier}): [{self.sourceitems}, {self.targetitems}]>"

    @property
    def issourcename(self) -> bool:
        """Return True if a single source term with pos=Name."""
        return len(self.sourceitems) == 1 and self.sourceitems[0].pos == "Name"

    def display(self) -> None:
        """Display text for alignments."""
        print(f"{self.identifier}: {[s.token for s in self.sourceitems]}\t{[s.token for s in self.targetitems]}")


def verseid(ag):
    """Key function for grouping by verse.

    Takes an alignment group and return the verse ID.
    """
    # not sure why this doesn't work
    # assert isinstance(ag, AlignmentGroup), f"Arg must be an AlignmentGroup instance: {ag}"
    assert ag.__class__.__name__ == "AlignmentGroup", f"Arg must be an AlignmentGroup instance: {ag}"
    return ag.identifier.split(".")[0]


@dataclass
class AlignmentSet:
    """Manage all the alignment data for a verse."""

    verseid: str
    sources: list = field(default_factory=list)
    targets: list = field(default_factory=list)
    items: list = field(default_factory=list)

    def __repr__(self) -> str:
        """Return a printed representation."""
        return f"<AlignmentSet({self.verseid})>"

    @staticmethod
    def from_aglist(aglist) -> "AlignmentSet":
        """Return an AlignmentSet for a list of AlignmentGroups."""
        assert aglist, f"List must not be empty: {aglist}"
        verseID = verseid(aglist[0])
        aset = AlignmentSet(verseid=verseID)
        aset.sources = [item for ag in aglist for item in ag.sourceitems]
        aset.targets = [item for ag in aglist for item in ag.targetitems]
        aset.items = aglist
        return aset

    # def namesets(self, sourceattr: str = "text") -> list:
    #     """Return a list of lists pairing a source name with target term(s).

    #     By default, the text attribute for the source is provided, or
    #     you can specify a different sourceattr.

    #     """
    #     return [
    #         (getattr(ag.sourceitems[0], sourceattr), t.text)
    #         for ag in self.items
    #         if ag.issourcename
    #         for t in ag.targetitems
    #     ]


class Reader(UserDict):
    """Read alignment data and integrate source and target."""

    def __init__(self, sourceid: str, targetid: str, languageid: str, processid: str) -> None:
        """Initialize Reader instance."""
        self.sourceid = sourceid
        self.targetid = targetid
        self.languageid = languageid
        self.processid = processid
        super().__init__(self)
        self.sourcereader = gcsource.Reader(sourceid, targetid)
        self.targetreader = gctarget.Reader(sourceid, targetid)
        # check here if a valid language
        # check here if a valid combination of sourceid, targetid, process
        alignmentdir = config.ALIGNMENTS / languageid / targetid
        # load all the alignments: not grouped by verse
        with (alignmentdir / f"{sourceid}-{targetid}-{processid}.json").open(encoding="utf-8") as f:
            self.data = {
                agid: AlignmentGroup(identifier=agid, sourceitems=sourceitems, targetitems=targetitems, meta=metadict)
                for aldict in json.load(f)
                # only a single key
                if (agid := list(aldict.keys())[0])
                # map the source identifiers to instances
                if (sourceitems := [self.sourcereader[s] for s in aldict[agid][sourceid]])
                # map the target identifiers to instances
                if (targetitems := [self.targetreader[t] for t in aldict[agid][targetid]])
                if (metadict := aldict[agid]["meta"])
            }

    def display(self) -> None:
        """Display configuration information for a Reader."""
        print(f"Source:\t{self.sourceid}\t({len(self.sourcereader)} words)")
        print(f"Target:\t{self.targetid}\t({len(self.targetreader)} words)")
        print(f"Process:\t{self.processid}")
        print(f"{len(self)} alignments")

    def source_concordance(self, value: str, attr: str = "lemma") -> list[AlignmentGroup]:
        """Return alignment groups whose source attr value is value."""
        return [ag for ag in self.data.values() if value in [getattr(s, attr) for s in ag.sourceitems]]

    def verse_groups(self) -> list[list[AlignmentGroup]]:
        """Return a list of lists of AlignmentGroups, by verse."""
        return [list(g) for k, g in itertools.groupby(self.values(), verseid)]
